#include <linux/module.h>
#include <linux/sched.h>
#include <linux/irq.h>
#include <linux/slab.h>
#include <linux/mm.h>
#include <asm/segment.h>
#include <asm/uaccess.h>
#include <asm/io.h>
#include <asm/delay.h>
#define SETTLE	50
#include "pc26at.h"

/* define which parms can be altered */

#define is_digit(c) ((c) >= '0' && (c) <= '9') 
#define PC26AT_MOD_NAME	"pc26at"
static int pc26at_major=PC26AT_MAJOR;
static int pc26at_base_address=PC26AT_BASE_ADDRESS;
static int pc26at_irq=PC26AT_IRQ;
/* don't bother with the DMA*/

MODULE_PARM(pc26at_major,"i");
MODULE_PARM(pc26at_base_address,"i");
MODULE_PARM(pc26at_irq,"i");
/*MODULE_AUTHOR("Matt Clark, Aug 2001");*/

static int pc26at_ioctl(struct inode *, struct file *, unsigned int , unsigned long);
static void pc26at_isr(int, void *, struct pt_regs*);
static ssize_t pc26at_read(struct file *, char *, size_t, loff_t *);
static ssize_t pc26at_write(struct file *, const char *, size_t, loff_t *);
static int pc26at_open(struct inode *, struct file *);
static int pc26at_release(struct inode *, struct file *);
static float pc26at_adconv(int); 
static int pc26at_handle_buffer(int);
static int pc26at_clockrate(int);

/* these data types occur here and are global to facilitate interrupt handling *
* since the a2d card is essentially a uniprocess board it makes sense to do this*/

PC26AT_BUFF_T * pc26at_data_buffer;
int	pc26at_data_pointer;
int	pc26at_buffer_size;
int	pc26at_channel_list[PC26AT_NO_CHANNELS];
int	pc26at_channel=0;
int	pc26at_request=0;
int	pc26at_no_channels=0;
int	pc26at_mode=PC26AT_TRIG_INT;


/* sleep queue for interrrupts */
DECLARE_WAIT_QUEUE_HEAD(pc26at_sleep);

#define TESTING	0


struct file_operations pc26at_fops = {
	read:		pc26at_read,
	write:		pc26at_write,
	ioctl:		pc26at_ioctl,
	open:		pc26at_open,
	release:	pc26at_release,
	};


static void pc26at_isr(int irq, void *dev_id,struct pt_regs *regs){
int top,bottom,val;
float volts;
// this is not a shared interrupt so don't bother to probe.
/* the board only generates the interrupt when it is ready	*
* so we don't have to wait for a conversion			*
* so now read values immediately				*/
	bottom=inb(pc26at_base_address+PC26AT_INPORT1);
	top=inb(pc26at_base_address+PC26AT_INPORT2)&PC26AT_IN2MSB;
	val=(top<<8) + bottom;
	if(pc26at_data_pointer==-1){
		pc26at_data_pointer++;
		return;
		}
	/* convert to volts */
	if (val < 2048) 	
		volts= (10.0 * val) /2047.0;
	else 
		volts=-(10.0*(4096-val))/2048.0;
	pc26at_data_buffer[pc26at_data_pointer++]=volts;
	/* if we have enough data wake up calling process 	*/
	/* and disable interrupts from this device iand return 	*/
	if(pc26at_data_pointer>=pc26at_request){
		/* prevent further interrupts from the board */
		//disable_irq(pc26at_irq);	
		/* disable the interrupt generation on the board */
		outb(0,pc26at_base_address+PC26AT_A2DCNTRL);
		/* awaken the process */
		wake_up(&pc26at_sleep);
		return;
		}

	/* else set the next channel (if required)		*/
	if(pc26at_no_channels!=1){
		outb(PC26AT_INTENB| PC26AT_DMAENB| ((char)pc26at_channel_list[pc26at_channel++]),pc26at_base_address+PC26AT_A2DCNTRL);
		if(pc26at_channel>=pc26at_no_channels)
			pc26at_channel=0;
		}
	}

static int pc26at_ioctl(struct inode * inode,struct file *file, unsigned int cmd, unsigned long arg){
	switch(cmd){
		case PC26AT_HARDRESET:
			/* resets the usage count to 1
			 for use after a module crash
			 make sure nothing else has opened
			 the module */
			printk("pc26at: ioctl: HARDRESET: attempting mod_use=1\n");
			/* pull it to 0 */
			while(MOD_IN_USE)MOD_DEC_USE_COUNT;
			MOD_INC_USE_COUNT;
			/* inc once for this open */
			printk("pc26at: ioctl: HARDRESET: Done\n");
			break;
		case PC26AT_TRIG_INT:
			pc26at_mode=PC26AT_TRIG_INT;
			break;
		/*case PC26AT_TRIG_SOFT:
			pc26at_mode=PC26AT_TRIG_SOFT;
			break;*/
		case PC26AT_HZ:
			return pc26at_clockrate(arg);
			break;
		default:
			printk("pc26at: ioctl: Unknown ioctl cmd\n");
			}
	return 0;
	}


int init_module(void) {
int res,inter;
	res=register_chrdev(pc26at_major,PC26AT_MOD_NAME,&pc26at_fops);
	if (res<0) return(res);

	inter=request_irq(pc26at_irq,pc26at_isr,0,"pc26at",NULL);
	if(inter<0) {
		// Can't get the requested interrupt
		printk("PC26AT: Can't register interrupt - quitting\n");
		unregister_chrdev(pc26at_major,PC26AT_MOD_NAME);
		return inter;
		}
	/* this card is tres stupid and if jumper can be interrupting like nobodies busines */
	/* discade all the interrupts until we want them */
	//disable_irq(pc26at_irq);
	pc26at_major=res;
	pc26at_no_channels=0;
	
	/* first set up the board correctly */
	outb(PC26AT_CONTROL_WORD,pc26at_base_address+PC26AT_BOARDCNTRL);

	/* announce us*/
	printk("pc26at dev: init_module: version 2.0 (c) Matt Clark Nov 2002- device %d\n",pc26at_major);
	return 0;
	} 

void cleanup_module(void){
int res;
	/* seriously need to think about releasing memory here.....*/
	/* it will be released during close()....		 */
	printk("pc26at dev: cleanup_module: Matt Clark aug 2001- device %d\n",pc26at_major);
	free_irq(pc26at_irq,NULL);
	res=unregister_chrdev(pc26at_major,PC26AT_MOD_NAME);
	} 


static ssize_t pc26at_read(struct file *file, char * buf,size_t count, loff_t *ppos) {
int i,request,temp;
unsigned long cnt;
	//printk("pc26at: in read\n");
	/* convert the numbers and return the buffer */
	/* drop any data longer than count */
	/* since the number of bytes will only be 0 */
	/* if the number of channels is 0 EOF will */
	/* never occur */
	
	/* work out how many conversions we want*/
	request=count/sizeof(PC26AT_BUFF_T);

	/* this is where I need to grow the buffer */

	temp=pc26at_handle_buffer(request);

	if(temp<0) 
		return temp;

	if(request>pc26at_buffer_size)
		request=pc26at_buffer_size;

	switch(pc26at_mode){
		/*case PC26AT_TRIG_SOFT:

			for(i=0,pc26at_channel=0;i<request;i++,pc26at_channel++){
				if(pc26at_channel>pc26at_no_channels)
					pc26at_channel=0;
				pc26at_data_buffer[i]=pc26at_adconv(pc26at_channel_list[pc26at_channel]);
				}

			cnt=copy_to_user(buf,(char *)pc26at_data_buffer,(unsigned long)(request*sizeof(PC26AT_BUFF_T)));

			if(cnt!=0){
				printk("pc26at: read: copy_to_user: ERROR trans (%lu) \n",cnt);
				return -EFAULT;
				}

			return request*sizeof(PC26AT_BUFF_T);

			break;
*/

		case PC26AT_TRIG_INT:
			/* set the first channel, pointer and request size */
			pc26at_channel=0;
			pc26at_data_pointer=-1;
			pc26at_request=request;
			/* the example used PC26AT_DMAENB as well without explanation	*
			 * but it doesn't seem necessary 				*/
			outb(PC26AT_INTENB | PC26AT_DMAENB |(char)pc26at_channel_list[pc26at_channel++],pc26at_base_address+PC26AT_A2DCNTRL);
                	if(pc26at_channel>=pc26at_no_channels)
                        	pc26at_channel=0;
			// read the first byte as a quick fix matt 2003 
			//inb(pc26at_base_address+PC26AT_INPORT1);
			//inb(pc26at_base_address+PC26AT_INPORT2)&PC26AT_IN2MSB;
					

			/* enable interrupts*/
			//enable_irq(pc26at_irq);
			/* sleep on queue */
			//printk("pc26at: read: About to sleep on queue (ints)\n");
			if(wait_event_interruptible(pc26at_sleep,(pc26at_data_pointer>=request))!=0)
				return -ERESTARTSYS;
			//printk("pc26at: read: woke up\n");
			//printk("pc26at: got %d bytes, going to transfer %d bytes\n",pc26at_data_pointer,request);
			/* transfer data */
			cnt=copy_to_user(buf,(char *)pc26at_data_buffer,(unsigned long)(request*sizeof(PC26AT_BUFF_T)));

			if(cnt!=0){
				printk("pc26at: read: copy_to_user: ERROR trans (%lu) \n",cnt);
				return -EFAULT;
				}
			return request*sizeof(PC26AT_BUFF_T);
				
		default:
			return -EINVAL;
		}			
	}

static ssize_t pc26at_write(struct file *file,const char * buf,size_t count, loff_t *ppos) {
int 	i;

	/* reprogram all a2d channels */
	if(count>PC26AT_NO_CHANNELS)return -E2BIG;
	pc26at_no_channels=count;

	/*interpret channels as bytes */
	for(i=0;i<count;i++){
		if((buf[i]>=PC26AT_NO_CHANNELS)||(buf[i]<0)){
			pc26at_no_channels=0;
			return -EINVAL;
			}
		pc26at_channel_list[i]=buf[i]<<4; /* note already converted nicely */
		}
	printk("pc26at: write: got %d channels\n",i);
	return count;
	}

static int pc26at_open(struct inode *inode, struct file *file) {
int temp;
	/* Should MOD_INC.. be here of after the possibility of returning with an error?*/
	MOD_INC_USE_COUNT;

	// NOV 2002 variable buffer size
	temp=pc26at_handle_buffer(PC26AT_DEF_BUFF);
	if(temp<0){
		printk("PC26AT: open: Fatal error: Unable to kmalloc for buffer\n");
		return -ENOMEM;
		}
		pc26at_no_channels=0;
	return 0;
	} 

static int pc26at_release(struct inode *inode, struct file *file) {
MOD_DEC_USE_COUNT;
return(0);
	}

static float pc26at_adconv(int channel){ 
int cnt=0,top,bottom,val;
float volts;
char byte;
	/* strobe the a2d cntrl for channel 0 */
	byte=PC26AT_SFTCLK|((char)channel);
	outb(byte,pc26at_base_address+PC26AT_A2DCNTRL);
	udelay(SETTLE);
	outb(byte+PC26AT_STROBE,pc26at_base_address+PC26AT_A2DCNTRL);
	outb(byte,pc26at_base_address+PC26AT_A2DCNTRL);
	/* now busy wait for the conversion; note DON'T STOP the kernel */
	while (((inb(pc26at_base_address+PC26AT_INPORT2) & PC26AT_BUSYIN)==0 )&&(cnt<100))cnt++; 
	/* now read values */
	bottom=inb(pc26at_base_address+PC26AT_INPORT1);
	top=inb(pc26at_base_address+PC26AT_INPORT2)&PC26AT_IN2MSB;
	val=(top<<8) + bottom;
	/* convert to volts */
	if (val < 2048)
	volts= (10.0 * val) /2047.0;
	else
		volts=-(10.0*(4096-val))/2048.0;
	volts=channel;
	return volts;
	}

static int pc26at_handle_buffer(int size){
	/* sanity checks */
	if(size<pc26at_buffer_size){
		//printk("pc26at: pc26at_handle_buffer: request to shrink buffer - ignored\n");
		return pc26at_buffer_size;
		}
	if(size>PC26AT_BUFF_MAX){
		printk("PC26AT: ioctl: Buffer size too big - not set\n");
		return -EINVAL;
		}
	/* now try to realloc the buffer memory */
	kfree(pc26at_data_buffer);
	pc26at_data_buffer=kmalloc(sizeof(PC26AT_BUFF_T)*size,GFP_KERNEL);
	if(pc26at_data_buffer==NULL){
		/* oh dear couldn't do that - try to get the original back */
		pc26at_data_buffer=kmalloc(sizeof(PC26AT_BUFF_T)*pc26at_buffer_size,GFP_KERNEL);
		printk("PC26AT: ioctl: could not enlarge buffer\n");
		if(pc26at_data_buffer==NULL){
			/* oh dear couldn't do that either - really in the shit now */
			pc26at_buffer_size=0;
			printk("PC26AT: ioctl: Error: unable to allocate any buffer\n");
			return -ENOMEM;
			}
		/* not sure here if to return -EMSGSIZE or buffersize*/
		return -EMSGSIZE;
		}
	else{
		pc26at_buffer_size=size;
		}
	/* success */
	/* not sure here if to return 0 or buffersize*/
	printk("pc26at: pc26at_handle_buffer: buffer grew to %d\n",pc26at_buffer_size);
	return pc26at_buffer_size;
	}

/* to use hardware interrupts you MUST have the board jumpered to use counter 0 alone */
static int pc26at_clockrate(int hz){
int period;
	if(hz<16)
		return	-EINVAL;
	period=(int)(float)PC26AT_MASTER_CLK/(float)hz;
        outb(0x36,pc26at_base_address+PC26AT_CNTCNTRL);
        outb((unsigned char) (period & 0xFF),pc26at_base_address+PC26AT_COUNTER0);
        outb((unsigned char) (period >> 8),pc26at_base_address+PC26AT_COUNTER0);
	return PC26AT_MASTER_CLK/period;
	}




