/* device driver for pc26at a2d card */
#include <linux/ioctl.h>
#ifndef PC26AT_MAJOR
#define PC26AT_MAJOR 0
#endif

/* following are the factory defaults for the PC26AT */
/* They are all acceptable for webface- matt aug 2001*/
#define PC26AT_BASE_ADDRESS	0X300
#define PC26AT_IRQ		7
#define PC26AT_DMA_LEVEL	1

#define PC26AT_MOD_NAME		"pc26at"

/* port defs */

/* A->D style */
#define PC26AT_PORTA	0
#define PC26AT_PORTB 	1
#define PC26AT_PORTC	2
#define PC26AT_PORTD	3
#define PC26AT_PORTE	4
#define PC26AT_PORTF	5
#define PC26AT_PORTG	6
#define PC26AT_PORTH	7

/* symbolic style */
#define PC26AT_INPORT1		0
#define PC26AT_INPORT2 	1
#define PC26AT_A2DCNTRL 	2
#define PC26AT_BOARDCNTRL 	3
#define PC26AT_COUNTER0	4
#define PC26AT_COUNTER1	5
#define PC26AT_COUNTER2	6
#define PC26AT_CNTCNTRL	7
#define PC26AT_MASTER_CLK	1000000
#define PC26AT_MAX_DIV		100

/* port vals */
/* PORTA / INPORT1	 input lsb */

/* PORTB / INPORT2	[0-3] input msb, [4] softtrigger, [5] ADC/BUSY, [6-7] reserved */
#define PC26AT_IN2MSB		0x0F
#define PC26AT_SOFTTRIG	0x10
#define PC26AT_BUSYIN		0x20

/* PORTC / A2DCNTRL	output [0] ADC +ve edge trig, [1] softclk (1) timer (0), [2] DMA enable, [3] int enable, [4-7] channel */
#define PC26AT_STROBE		0x01
#define PC26AT_SFTCLK		0x02
#define PC26AT_DMAENB		0x04
#define PC26AT_INTENB		0x08
#define PC26AT_CHNMSK		0xF0 /* mask for channels */
#define PC26AT_CHNMLT		0x0F /* multiplier for channels */

/* PORTD / BOARDCNTL */
#define PC26AT_CONTROL_WORD	0x92	/* defined in manual- must be set*/

/* now onto internals */

/* the number of channels for this board */
#define PC26AT_NO_CHANNELS 16

/* ioctl numbers */
#define PC26AT_MAGIC		100
/* make it possible to remove the module if crashed */
#define PC26AT_HARDRESET	_IO(PC26AT_MAGIC,0)
#define PC26AT_TRIG_INT		_IO(PC26AT_MAGIC,1)
//#define PC26AT_TRIG_SOFT	_IO(PC26AT_MAGIC,2)
#define PC26AT_HZ		_IO(PC26AT_MAGIC,3)

#define PC26AT_DEF_BUFF		16*4 

#define PC26AT_BUFF_T		float /* stick with floats */
#define PC26AT_BUFF_MAX		1024*128/sizeof(PC26AT_BUFF_T)


/* since the driver is single open by design we will forego 	*
 * private data and use global data instead which makes 	*
 * handling slightly easier - since the card is single threaded *
 * it doesn't hurt to do this (beyond having a little extra	*
 * kernel memory allocated while the driver is closed matt 2002	*/
